package com.ksyun.util;

import com.fasterxml.jackson.annotation.JsonInclude;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.common.collect.Lists;
import com.ksyun.constant.Consts;
import jodd.util.StringUtil;
import org.apache.commons.codec.digest.HmacAlgorithms;
import org.apache.commons.codec.digest.HmacUtils;
import org.joda.time.format.DateTimeFormat;
import org.joda.time.format.DateTimeFormatter;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class UtilAll {

    private UtilAll() {

    }

    private static final Logger logger = LoggerFactory.getLogger(UtilAll.class);

    private static final DateTimeFormatter timeFormatter = DateTimeFormat.forPattern("yyyy-MM-dd'T'HH:mm:ss'Z'").withZoneUTC();

    private static final String DEFAULT_ENCODING = "UTF-8";

    private static final ObjectMapper mapper = new ObjectMapper();


    private static final Pattern ENCODED_CHARACTERS_PATTERN;

    static {
        StringBuilder pattern = new StringBuilder();

        pattern
                .append(Pattern.quote("+"))
                .append("|")
                .append(Pattern.quote("*"))
                .append("|")
                .append(Pattern.quote("%7E"))
                .append("|")
                .append(Pattern.quote("%2F"));

        ENCODED_CHARACTERS_PATTERN = Pattern.compile(pattern.toString());
    }

    static {
        mapper.setSerializationInclusion(JsonInclude.Include.NON_NULL);
        mapper.disable(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES);
    }

    public static boolean isNullOrEmpty(String string) {
        return string == null || string.isEmpty();
    }

    public static String encode(Map<String, String> params) {
        Map<String, String> map = new HashMap<>(params);

        List<String> keys = new ArrayList<>(map.keySet());
        Collections.sort(keys);
        if (map == null || map.size() == 0) {
            return "";
        }

        String body = "";
        for (String key : keys) {
            String value = map.get(key);
            body = body + key + '=' + encode(value, false) + '&';
        }
        return body.substring(0, body.length() - 1);
    }

    private static String encode(final String value, final boolean path) {
        if (value == null) {
            return "";
        }
        try {
            String encoded = URLEncoder.encode(value, DEFAULT_ENCODING);
            Matcher matcher = ENCODED_CHARACTERS_PATTERN.matcher(encoded);
            StringBuffer buffer = new StringBuffer(encoded.length());
            while (matcher.find()) {
                String replacement = matcher.group(0);

                if ("+".equals(replacement)) {
                    replacement = "%20";
                } else if ("*".equals(replacement)) {
                    replacement = "%2A";
                } else if ("%7E".equals(replacement)) {
                    replacement = "~";
                } else if (path && "%2F".equals(replacement)) {
                    replacement = "/";
                }

                matcher.appendReplacement(buffer, replacement);
            }
            matcher.appendTail(buffer);
            return buffer.toString();
        } catch (UnsupportedEncodingException ex) {
            throw new RuntimeException(ex);
        }
    }

    public static Map<String, String> sign(String ak, String sk, String region, String action, Map<String, String> privateParams) {
        Map<String, String> params = new LinkedHashMap<>();
        params.put("Accesskey", ak);
        if (Consts.smsServiceAction.contains(action)) {
            params.put("Service", Consts.SMC_SERVICE);
        } else {
            params.put("Service", Consts.KSMC_SERVICE);
        }
        params.put("Action", action);
        params.put("Version", Consts.Version);
        params.put("Timestamp", timeFormatter.print(System.currentTimeMillis()));
        params.put("SignatureVersion", "1.0");
        params.put("SignatureMethod", "HMAC-SHA256");
        params.put("Region", region);
        params.putAll(privateParams);
        String encodeParam = UtilAll.encode(params);
        logger.info("encodeParam:{} ", encodeParam);
        String signature = new HmacUtils(HmacAlgorithms.HMAC_SHA_256, sk).hmacHex(encodeParam);
        params.put("Signature", signature);
        logger.info("allParam:{}", UtilAll.encode(params));
        return params;
    }

    /**
     * Object可以是POJO，也可以是Collection或数组。 如果对象为Null, 返回"null". 如果集合为空集合, 返回"[]".
     */
    public static String toJson(Object object) {
        try {
            return mapper.writeValueAsString(object);
        } catch (IOException e) {
            throw new RuntimeException("write to json string error:" + Objects.toString(object), e);
        }
    }

    /**
     * 反序列化POJO或简单Collection如List<String>.
     * <p>
     * 如果JSON字符串为Null或"null"字符串, 返回Null. 如果JSON字符串为"[]", 返回空集合.
     * <p>
     * 如需反序列化复杂Collection如List<MyBean>, 请使用fromJson(String, JavaType)
     */
    public static <T> T fromJson(String jsonString, Class<T> clazz) {
        try {
            return mapper.readValue(jsonString, clazz);
        } catch (IOException e) {
            throw new RuntimeException("parse json string error:" + jsonString, e);
        }
    }

    public static <T> T fromJson(String jsonString, TypeReference valueTypeRef) {
        if (jsonString == null || jsonString.trim().length() == 0) {
            return null;
        }
        try {
            return (T) mapper.readValue(jsonString, valueTypeRef);
        } catch (IOException e) {
            logger.error("parse json string error:" + jsonString, e);
            return null;
        }
    }

    public static Map<String, String> buildMap(String... values) {
        List<String> lists = Arrays.asList(values);
        if (lists.size() % 2 == 1) {
            throw new IllegalArgumentException("参数个数须为偶数个");
        }
        List<List<String>> listParts = Lists.partition(lists, 2);
        Map<String, String> params = new HashMap<>();
        for (List<String> kv : listParts) {
            if (StringUtil.isNotBlank(kv.get(1))) {
                params.put(kv.get(0), kv.get(1));
            }
        }
        return params;
    }

}
